/*
 * Logdevice - A device used to record debuging information in the kernel.
 *             It uses a large memory ring buffer consisting of individual pages
 *             to keep down on hogging large sections. A user may then read the device
 *             to get debugging information out of it.  Or if configured, this can
 *             be dumped to the network on a system crash.
 *
 * Copyright - 2005 - Steven Rostedt, Kihon Technologies, (rostedt at kihontech dot com)
 */
#ifndef _LOG_DEV_H
#define _LOG_DEV_H

#define LOGDEV_CUSTOM 0x1afb
#define LOGDEV_SWITCH_ID 0x2afc
#define LOGDEV_PKT_ID 0x42aa
#define LOGDEV_PRINT 0x4adb
#define LOGDEV_PRINT_TIME 0x4adc
#define LOGDEV_PRINT_TIME_FUNC 0x4add

struct logdev_switch_struct {
	unsigned long high;
	unsigned long low;
	short pid_prev;
	short pid_next;
	short prev_len;
	short next_len;
	unsigned long wchan;
	char prev_comm[0];
	char next_comm[0];
};

struct logdev_pkt {
	unsigned long high;
	unsigned long low;
	short protocol;
	short dir;
	char packet[0];
};

struct logdev_print {
	char str[0];
};

struct logdev_print_time {
	unsigned long high;
	unsigned long low;
	char str[0];
};

struct logdev_print_time_func {
	const char *file;
	int line;
	/* need to be after line, since we use this with print_time. */
	unsigned long high;
	unsigned long low;
	char str[0];
};

struct logdev_custom {
	int id;
	char data[0];
};

struct logdev_header {
	int id;
	int size;
	int cpu;
};

struct logdev_item {
	struct logdev_header hdr;
	union {
		struct logdev_switch_struct sw;
		struct logdev_pkt pkt;
		struct logdev_print print;
		struct logdev_print_time print_time;
		struct logdev_print_time_func print_time_func;
		struct logdev_custom custom;
		char data[0];
	} u;
};

#ifdef __KERNEL__
#include <linux/config.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <asm/atomic.h>

/*
 * I also use this with Ingo Molnar's RT kernel.
 */
#ifndef CONFIG_PREEMPT_RT
#undef raw_spinlock_t
#undef RAW_SPIN_LOCK_UNLOCKED
#define raw_spinlock_t spinlock_t
#define RAW_SPIN_LOCK_UNLOCKED SPIN_LOCK_UNLOCKED
#define spin_lock_rt do {} while(0)
#endif

#ifdef CONFIG_LOGDEV_HOOKS
extern int (*logdev_print_hook)(const char *fmt, ...);
extern int (*logdev_vprint_hook)(const char *fmt, va_list va);
extern int (*logdev_print_time_hook)(const char *fmt, ...);
extern int (*logdev_print_time_func_hook)(const char *fmt, ...);
extern void (*logdev_record_switch_hook)(struct task_struct *a,struct task_struct *b);
extern int (*logdev_record_hook)(int id, int size, const void *data);
extern int (*logdev_write_hook)(const char *data, int size);
extern int (*logdev_read_hook)(void *buf, int size);
extern void (*logdev_pkt_hook)(struct sk_buff *skb, int dir);
extern raw_spinlock_t logdev_add_hook_lock;
extern void (*logdev_dump_hook)(void);
extern void (*logdev_dumpnet_hook)(void);

extern atomic_t logdev_level;
extern atomic_t logdev_switch;
extern atomic_t logdev_print_enabled;

#define logdev_switch_on() atomic_inc(&logdev_switch)
#define logdev_switch_off() \
	do { \
		if (atomic_read(&logdev_switch)) \
			atomic_dec(&logdev_switch); \
	} while(0);
void logdev_write_msg(const char *msg, unsigned int len, int net);

typedef void (*logdev_callback_func)(struct logdev_header *hdr, 
				     struct logdev_custom *custom,
				     int net,
				     void *rec);

/*
 * Only exists when compiled in.
 */
#ifdef CONFIG_LOGDEV
void logdev_gettimeofday(struct timeval *tv);
#endif

#ifndef CONFIG_LOGDEV
/*
 * dumpnet is expected to be called on a bug, so we don't
 * need to lock it.
 */
#define LOGDUMP() do { if (logdev_dump_hook) logdev_dump_hook(); } while(0)
#define LOGDUMPNET() do { if (logdev_dumpnet_hook) logdev_dumpnet_hook(); } while(0)

#define LOGDEV(x,y...) do { \
	spin_lock(&logdev_add_hook_lock); \
	if (logdev_##x##_hook) \
		logdev_##x##_hook(y); \
	spin_unlock(&logdev_add_hook_lock); \
} while (0)
#else

void logdev_record_switch(struct task_struct *prev, struct task_struct *next);
void logdev_pkt(struct sk_buff *skb, int direction);
int logdev_print(const char *str, ...)
	__attribute__ ((format (printf, 1, 2)));
int logdev_vprint(const char *str, va_list va);
int logdev_print_time(const char *str, ...)
	__attribute__ ((format (printf, 1, 2)));
int logdev_print_time_func(const char *file, int line, const char *str, ...)
	__attribute__ ((format (printf, 3, 4)));
int logdev_record(int id, int size, const void *data);
int logdev_record_write(const char *data, int size);
int logdev_record_read(void *data, int size);
void logdev_dumpnet(void);
void logdev_dump(void);
int logdev_register_callback(int custom_id, logdev_callback_func func);
int logdev_unregister_callback(int custom_id);

#define LOGDUMP() logdev_dump()
#define LOGDUMPNET() logdev_dumpnet()
#define LOGDEV(x,y...) logdev_##x(y)

#endif
#define LOGPRINT(x...) do { if (atomic_read(&logdev_print_enabled)) LOGDEV(print,x); } while(0)
#define LOGTPRINT(x...) do { if (atomic_read(&logdev_print_enabled)) LOGDEV(print_time,x); } while(0)
#define LOGTFPRINT(x...) do { if (atomic_read(&logdev_print_enabled)) LOGDEV(print_time_func,__FUNCTION__,__LINE__,x); } while(0)
#define LOGSWITCH(prev,next) LOGDEV(record_switch,prev,next)
#define LOGPKT(skb,direction) LOGDEV(pkt,skb,direction)
void logdev_time(void);

#define lprint(x...) do { if (atomic_read(&logdev_switch)) { LOGTPRINT(x); } } while(0)

#else /* !LOGDEV_HOOKS */
#define LOGDUMP() do {} while(0)
#define LOGDEV(x,y...) do {} while(0)
#define LOGSWITCH(prev,next) do {} while(0)
#define LOGPKT(skb,direction) do {} while(0)
#define logdev_switch_on() do {} while(0)
#define logdev_switch_off() do {} while(0)
#define logdev_time() do {} while(0)
#endif /* LOGDEV_HOOKS */

#endif

extern int logdev_debug;
#if 0
#undef LOGPRINT
#undef LOGSWITCH
#define LOGPRINT printk
#define LOGSWITCH(prev,next) do { if (atomic_read(&logdev_switch)) printk("(%d:%s)-->(%d:%s)\n",(prev)->pid,(prev)->comm,(next)->pid,(next)->comm); } while(0)
#endif
extern void __edprint(int cpu, const char *func, int line, const char *fmt, ...)
	__attribute__ ((format (printf, 4, 5)));
extern void logdev_check_stack(void);
#if 0
#define edprint(x...) do { logdev_check_stack(); if (atomic_read(&logdev_switch)) { \
	__edprint(_smp_processor_id(),__FUNCTION__,__LINE__,x);}} while(0)
#else
#define _edprint(x,y...) do { logdev_check_stack(); if (atomic_read(&logdev_switch)) { \
    LOGTFPRINT(x "%s\n", y);} } while(0)
#define edprint(x...) _edprint(x,"")
#endif


#endif
